#include "light.h"
#include "framebuffer.h"
#include "texture.h"
#include "scene.h"
#include "shader.h"
#include "shadersource.h"
#include "phongmaterial.h"
#include "mesh.h"
#include <resource.h>
#include <glm/ext.hpp>

RESOURCE_PACK(shaders)

const char* Light::flagStr[] = {
    "AMBIENT_LIGHT",
    "DIRECTIONNAL_LIGHT",
    "POINT_LIGHT",
    "SPOT_LIGHT",
    "SHADOWMAP"
};

const glm::mat4 Light::biasMatrix(
        0.5, 0.0, 0.0, 0.0,
        0.0, 0.5, 0.0, 0.0,
        0.0, 0.0, 0.5, 0.0,
        0.5, 0.5, 0.5, 1.0);

Light::Light()
{
    initPointLight();
}

void Light::initAmbientLight(glm::vec3 lightColor)
{
    type = AMBIENT;
    position = glm::vec3(0);
    color = lightColor;
    shadowCaster = false;
    isDir = false;
    viewMatrix = glm::mat4();
}

void Light::initDirectionnalLight(glm::vec3 dir, glm::vec3 lightColor)
{
    type = DIRECTIONNAL;
    position = glm::vec3(0);
    direction = dir;
    color = lightColor;
    shadowCaster = false;
    isDir = true;
    viewMatrix = glm::lookAt(position, position+direction, glm::vec3(0, 1, 0));
}

void Light::initPointLight(glm::vec3 pos, glm::vec3 lightColor, float att)
{
    type = POINT;
    position = pos;
    cutOffAngle = 360;
    color = lightColor;
    shadowCaster = false;
    isDir = false;
    viewMatrix = glm::mat4();
    attenuation = att;
}

void Light::initSpotLight(glm::vec3 pos, glm::vec3 dir, float spotAngle, glm::vec3 lightColor)
{
    type = SPOT;
    position = pos;
    direction = dir;
    cutOffAngle = spotAngle;
    color = lightColor;
    shadowCaster = false;
    isDir = false;
}

bool Light::isDirectionnal()
{
    return isDir;
}

void Light::initShadowMap(int resWidth, int resHeight, glm::vec3 dim)
{
    shadowMapWidth = resWidth;
    shadowMapHeight = resHeight;
    viewMatrix = glm::lookAt(position, position-direction, glm::vec3(0, 1, 0));
    if(type == DIRECTIONNAL)
        projectionMatrix = glm::ortho(-dim.x/2, dim.x/2, -dim.y/2, dim.y/2, -dim.z/2, dim.z/2);
    else
        ; // TODO : glm::projection

    shadowCaster = true;
    shadowMap = new FrameBuffer();
    // Depth buffer
    Texture* tex = new Texture(GL_DEPTH_COMPONENT, GL_DEPTH_COMPONENT, resWidth, resHeight, GL_FLOAT);
    tex->setFiltering(GL_LINEAR);
    tex->setWrap(GL_CLAMP_TO_EDGE);
    tex->setParameter(GL_TEXTURE_COMPARE_MODE, GL_COMPARE_REF_TO_TEXTURE);
    tex->setParameter(GL_TEXTURE_COMPARE_FUNC, GL_LESS);
    shadowMap->addTexture(tex, GL_DEPTH_ATTACHMENT);
    shadowMap->initColorAttachments();

    ShaderSource source;
    Resource::ResourceMap shaderMap;
    Resource::getResourcePack_shaders(shaderMap);
    source.setSource(shaderMap["shaders/shadow.vert.glsl"], ShaderSource::VERTEX);
    source.setSource(shaderMap["shaders/shadow.frag.glsl"], ShaderSource::FRAGMENT);
    shaders[0] = source.compile(Mesh::MESH_3D, getFlags());
    shaders[1] = source.compile(Mesh::MESH_3D & Mesh::MATERIAL_ALPHA_MASK, getFlags());
}

void Light::generateShadowMap(Scene* scene)
{
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glViewport(0, 0, shadowMapWidth, shadowMapHeight);
    shadowMap->bindFBO();
    glClearDepth(1.0);
    glClear(GL_DEPTH_BUFFER_BIT);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    //glCullFace(GL_FRONT);
    for(SceneIterator<GeometryNode*>* geometryIt = scene->getGeometry();
        geometryIt->isValid(); geometryIt->next())
    {
        GeometryNode* node = geometryIt->getItem();
        if(node->mesh->getFlags() & (1 << Mesh::MESH_SHADOWED))
        {
            // compute matrix attributes
            glm::mat4 lightMVP = getProjectionMatrix() * (getViewMatrix() * node->modelMatrix);
            int hasAlpha = (node->mesh->getFlags() & (1 << Mesh::MATERIAL_ALPHA_MASK)) > 0;
            shaders[hasAlpha]->bind();
            shaders[hasAlpha]->bindMat4(shaders[hasAlpha]->getLocation("MVP"), lightMVP);
            node->mesh->draw(shaders[hasAlpha], false, hasAlpha, false);
        }
    }
    //glCullFace(GL_BACK);
}

Texture* Light::getShadowMap()
{
    return shadowMap->getTexture(0);
}

void Light::setPosition(glm::vec3 new_pos)
{
    position = new_pos;
    viewMatrix = glm::lookAt(position, position+direction, glm::vec3(0, 1, 0));
}

unsigned int Light::getFlags()
{
    unsigned int flags = 0;
    switch(type)
    {
    case AMBIENT :
        flags = 1 << AMBIENT_FLAG;
        break;
    case DIRECTIONNAL :
        flags = 1 << DIRECTIONNAL_FLAG;
        break;
    case POINT :
        flags = 1 << POINT_FLAG;
        break;
    case SPOT :
        flags = 1 << SPOT_FLAG;
        break;
    }
    if(shadowCaster)
        flags |= 1 << SHADOWMAP_FLAG;
    return flags;
}
