#include "mesh.h"
#include <glm/ext.hpp>
#include <set>

#include "sparrowrenderer.h"
#include "material.h"
#include "buffer.h"

const char* const Mesh::flagStr[Mesh::NB_FLAGS] =
{
    "INDEXED",
    "TEXTURABLE",
    "INSTANCED",

    "MESH_3D",
    "MESH_2D",

    "TANGENT_SPACE",
    "BILLBOARD",
    "SHADOWED",

    "COLOR_TEXTURE",
    "ALPHA_MASK",

    "PHONG",
    "DIFFUSE_TEXTURE",
    "AMBIENT_TEXTURE",
    "SPECULAR_TEXTURE",
    "NORMAL_MAP",

    "BUMP_MAP"
};

Mesh::Mesh() :
    material(NULL),
    isDoubleSided(false),
    isBillboard(false),
    isShadowCaster(true),
    depth(0),
    primitive_type(GL_TRIANGLES),
    vao(0)
{
    clearBuffers();
}

Mesh::~Mesh()
{
    destroyGL();
}

void Mesh::addBuffer(Buffer *b, int bufferType)
{
    buffersId[bufferType] = buffers.size();
    buffers.push_back(b);
}

void Mesh::clearBuffers()
{
    for(int i=0; i<NB_BUFFERS; ++i)
        buffersId[i] = -1;
    for(Buffer *b : buffers)
        delete(b);
    buffers.clear();
}

void Mesh::initGL()
{
    destroyGL();
    
    // create VAO
    glGenVertexArrays(1, &vao);
    glBindVertexArray(vao);

    // init positions VBO
    if(!positions3D.empty())
    {
        auto posBuffer = new TBuffer<glm::vec3>(positions3D, Buffer::VBO);
        posBuffer->setVertexAttrib(0, 3);
        addBuffer(posBuffer, POSITION_BUFFER);
        
        // init normals vbo
        if(!normals.empty())
        {
            auto normalsBuffer = new TBuffer<glm::vec3>(normals, Buffer::VBO);
            normalsBuffer->setVertexAttrib(1, 3);
            addBuffer(normalsBuffer, NORMAL_BUFFER);
        }
        // init tangents vbo
        if(!tangents.empty())
        {
            auto tangentsBuffer = new TBuffer<Tangents>(tangents, Buffer::VBO);
            tangentsBuffer->setVertexAttrib(3, 3, 0);
            tangentsBuffer->setVertexAttrib(4, 3, sizeof(glm::vec3));
            addBuffer(tangentsBuffer, TANGENT_BUFFER);
        }
    }
    else if(!positions2D.empty())
    {
        auto posBuffer = new TBuffer<glm::vec2>(positions2D, Buffer::VBO);
        posBuffer->setVertexAttrib(0, 2);
        addBuffer(posBuffer, POSITION_BUFFER);
    }
    else
    {
        fprintf(stderr, "ERROR : Mesh can't be initialized without position data");
        return;
    }
    // init texCoords vbo
    if(!texCoords.empty())
    {
        auto texCoordBuffer = new TBuffer<glm::vec2>(texCoords, Buffer::VBO);
        texCoordBuffer->setVertexAttrib(2, 2);
        addBuffer(texCoordBuffer, TEXCOORD_BUFFER);
    }
    
    // init instances vbo
    if(!instances_offsets.empty())
    {
        auto instancesBuffer = new TBuffer<glm::vec3>(instances_offsets, Buffer::VBO);
        instancesBuffer->setVertexAttrib(5, 3, 0, 1);
        addBuffer(instancesBuffer, INSTANCE_BUFFER);
    }
        
        // init EBO
    if(!indices.empty())
    {
        auto indicesBuffer = new TBuffer<GLuint>(indices, Buffer::EBO);
        addBuffer(indicesBuffer, INDICES_BUFFER);
    }

    // unbind vao
    glBindVertexArray(0);
}

void Mesh::draw(Shader* shader, bool drawNormals, bool drawTexCoord, bool drawTangents)
{
    if(isDoubleSided)
        glDisable(GL_CULL_FACE);
    if(material != NULL)
        material->bindAttributes(shader);
    glBindVertexArray(vao);
    
    if(indices.empty())
    {
        int size = positions3D.empty() ? positions2D.size() : positions3D.size();
        if(!instances_offsets.empty())
        {
            glDrawArraysInstanced(primitive_type, 0, size, instances_offsets.size());
        }
        else
        {
            glDrawArrays(primitive_type, 0, size);
        }
    }
    else
    {
        Buffer *b = buffers[buffersId[INDICES_BUFFER]];
        b->bind();
        if(!instances_offsets.empty())
        {
            glDrawElementsInstanced(primitive_type, indices.size(), GL_UNSIGNED_INT, NULL, instances_offsets.size());
        }
        else
        {
            glDrawElements(primitive_type, indices.size(), GL_UNSIGNED_INT, NULL);
        }
        b->unbind();
    }
    
    glBindVertexArray(0);
    if(isDoubleSided)
        glEnable(GL_CULL_FACE);
}

void Mesh::destroyGL()
{
    if(vao != 0)
    {
        glDeleteVertexArrays(1, &vao);
        vao = 0;
    }
    clearBuffers();
}

unsigned int Mesh::getFlags()
{
    unsigned int flags = material->getFlags();

    if(!indices.empty())
        flags |= 1 << MESH_INDEXED;
    if(!texCoords.empty())
        flags |= 1 << MESH_TEXTURABLE;
    if(!instances_offsets.empty())
        flags |= 1 << MESH_INSTANCED;

    if(!positions3D.empty())
    {
        flags |= 1 << MESH_3D;

        if(!tangents.empty())
            flags |= 1 << MESH_TANGENT_SPACE;
        if(isBillboard)
            flags |= 1 << MESH_BILLBOARD;
        if(isShadowCaster)
            flags |= 1 << MESH_SHADOWED;
    }
    else
        flags |= 1 << MESH_2D;

    return flags;
}

struct VertexComparator
{
    // c'est plutot crade mais j'ai pas trouve d'autre moyen pour le moment
    static Mesh* mesh;
    static void setMesh(Mesh* m) {VertexComparator::mesh = m;}

    bool operator() (const int& vertId1, const int& vertId2) const
    {
        if(mesh->positions3D[vertId1].x != mesh->positions3D[vertId2].x)
            return (mesh->positions3D[vertId1].x < mesh->positions3D[vertId2].x);
        if(mesh->positions3D[vertId1].y != mesh->positions3D[vertId2].y)
            return (mesh->positions3D[vertId1].y < mesh->positions3D[vertId2].y);
        if(mesh->positions3D[vertId1].z != mesh->positions3D[vertId2].z)
            return (mesh->positions3D[vertId1].z < mesh->positions3D[vertId2].z);
        if(!mesh->texCoords.empty())
        {
            if(mesh->texCoords[vertId1].x - floor(mesh->texCoords[vertId1].x) != mesh->texCoords[vertId2].x - floor(mesh->texCoords[vertId2].x))
                return (mesh->texCoords[vertId1].x < mesh->texCoords[vertId2].x);
            if(mesh->texCoords[vertId1].y - floor(mesh->texCoords[vertId1].y) != mesh->texCoords[vertId2].y - floor(mesh->texCoords[vertId2].y))
                return (mesh->texCoords[vertId1].y < mesh->texCoords[vertId2].y);
        }
        if(!mesh->normals.empty())
        {
            if(mesh->normals[vertId1].x != mesh->normals[vertId2].x)
                return (mesh->normals[vertId1].x < mesh->normals[vertId2].x);
            if(mesh->normals[vertId1].y != mesh->normals[vertId2].y)
                return (mesh->normals[vertId1].y < mesh->normals[vertId2].y);
            if(mesh->normals[vertId1].z != mesh->normals[vertId2].z)
                return (mesh->normals[vertId1].z < mesh->normals[vertId2].z);
        }
        return false;
    }
};

Mesh* VertexComparator::mesh = NULL;

void Mesh::mergeVertices()
{
    if(positions3D.empty())
        return;
    bool *deleted = new bool[positions3D.size()];
    int *offsets = new int[positions3D.size()];
    std::set<int, VertexComparator> vertexSet;
    VertexComparator::setMesh(this);

    for(std::size_t i=0; i<indices.size(); ++i)
    {
        std::pair<std::set<int,VertexComparator>::iterator, bool> ret = vertexSet.insert(indices[i]);
        std::set<int,VertexComparator>::iterator it = ret.first;
        bool success = ret.second;
        deleted[indices[i]] = !success && *it != int(indices[i]);
        if(deleted[indices[i]])
        {
            if(!tangents.empty())
                tangents[*it].tangent += tangents[indices[i]].tangent;
            indices[i] = *it;
        }
    }
    int offset = 0;
    int pos = 0;
    for(std::size_t i=0; i<positions3D.size(); ++i)
    {
        if(deleted[i])
            ++offset;
        else
        {
            offsets[i] = offset;
            if(offset != 0)
            {
                positions3D[pos] = positions3D[i];
                if(!texCoords.empty())
                    texCoords[pos] = texCoords[i];
                if(!normals.empty())
                    normals[pos] = normals[i];
                if(!tangents.empty())
                    tangents[pos] = tangents[i];
            }
            ++pos;
        }
    }
    for(std::size_t i=0; i<indices.size(); ++i)
        indices[i] -= offsets[indices[i]];

    positions3D.resize(positions3D.size()-offset);
    if(!texCoords.empty())
        texCoords.resize(texCoords.size()-offset);
    if(!normals.empty())
        normals.resize(normals.size()-offset);
    if(!tangents.empty())
        tangents.resize(tangents.size()-offset);
    for(int i=0; i<int(tangents.size()); ++i)
    {
        glm::vec3 &T = tangents[i].tangent;
        const glm::vec3 &N = normals[i];
        
        T = glm::normalize(T);
        // re-orthogonalize T with respect to N
        T = glm::normalize(T - glm::dot(T, N) * N);
        // then retrieve perpendicular vector B with the cross product of T and N
        tangents[i].binormal = glm::normalize(glm::cross(T, N));
    }
}

void Mesh::computeNeighbors()
{
    if(positions3D.empty())
        return;
    // TODO : compute adjacency and change primitivetype
}

void Mesh::computeNormals()
{
    if(positions3D.empty())
        return;
    normals.resize(positions3D.size());
    std::memset(normals.data(), 0, normals.size());
    for (std::size_t i=0; i < indices.size(); i += 3)
    {
        int v0 = indices[i];
        int v1 = indices[i+1];
        int v2 = indices[i+2];
        glm::vec3 n = glm::cross(positions3D[v1] - positions3D[v0], positions3D[v2] - positions3D[v0]);
        normals[v0] += n;
        normals[v1] += n;
        normals[v2] += n;
    }
    for(glm::vec3 &n : normals)
        n = glm::normalize(n);
}

void Mesh::computeTangents()
{
    if(texCoords.empty())
        return;
    tangents = std::vector<Tangents>(positions3D.size());

    for (std::size_t j=0; j < indices.size(); j += 3)
    {
        int vertexId0 = indices[j];
        int vertexId1 = indices[j+1];
        int vertexId2 = indices[j+2];

        const glm::vec3 &v1 = positions3D[vertexId0];
        glm::vec3 edge1 = positions3D[vertexId1] - v1;
        glm::vec3 edge2 = positions3D[vertexId2] - v1;
        
        const glm::vec2& w1 = texCoords[vertexId0];
        glm::vec2 deltaUV1 = texCoords[vertexId1] - w1;
        glm::vec2 deltaUV2 = texCoords[vertexId2] - w1;
        
        float f = 1.0f / (deltaUV1.x * deltaUV2.y - deltaUV2.x * deltaUV1.y);
        
        glm::vec3 tangentDir = deltaUV2.y * edge1 - deltaUV1.y * edge2;
        tangentDir = glm::normalize(tangentDir*f);
        
        glm::vec3 binormalDir = deltaUV1.x * edge2 - deltaUV2.x * edge1;
        binormalDir = glm::normalize(binormalDir*f);
        
        tangents[vertexId0] = {tangentDir, binormalDir};
        tangents[vertexId1] = {tangentDir, binormalDir};
        tangents[vertexId2] = {tangentDir, binormalDir};
    }
}

void Mesh::computeBoundingBox(glm::vec3 &min, glm::vec3 &max)
{
    min = max = positions3D[0];
    for(const glm::vec3 &pos : positions3D)
    {
        if(pos.x < min.x)
            min.x = pos.x;
        else if(pos.x > max.x)
            max.x = pos.x;
        if(pos.y < min.y)
            min.y = pos.y;
        else if(pos.y > max.y)
            max.y = pos.y;
        if(pos.z < min.z)
            min.z = pos.z;
        else if(pos.z > max.z)
            max.z = pos.z;
    }
}
