#include "texture.h"

#include "image.h"

Texture::Texture(GLenum format,
                 GLenum internal_format,
                 int width,
                 int height,
                 GLenum dataType,
                 GLenum texTarget) :
    m_texSlot(0),
    m_target(texTarget),
    m_format(format),
    m_internal_format(internal_format),
    m_width(width),
    m_height(height),
    m_dataType(dataType),
    m_hasMipMaps(false)
{
    glGenTextures(1, &m_texId);
    glBindTexture(m_target, m_texId);
    switch(m_target)
    {
    case GL_TEXTURE_2D :
        glTexImage2D(m_target, 0, internal_format, width, height, 0, format, dataType, NULL);
        setWrap(GL_REPEAT);
        setFiltering(GL_LINEAR);
        break;
    case GL_TEXTURE_RECTANGLE :
        glTexImage2D(m_target, 0, internal_format, width, height, 0, format, dataType, NULL);
        setWrap(GL_CLAMP_TO_EDGE);
        setFiltering(GL_LINEAR);
        break;
    case GL_TEXTURE_CUBE_MAP :
        for(int i=0; i<6; ++i)
            glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, internal_format, width, height, 0, format, dataType, NULL);
        setWrap(GL_CLAMP_TO_EDGE);
        setFiltering(GL_LINEAR);
        break;
    }
}

Texture::Texture(Image* myImage, bool makeMipMaps) :
    m_texSlot(0),
    m_target(GL_TEXTURE_2D),
    m_width(myImage->width),
    m_height(myImage->height),
    m_dataType(GL_UNSIGNED_BYTE),
    m_hasMipMaps(makeMipMaps)
{
    glGenTextures(1, &m_texId);
    glBindTexture(m_target, m_texId);
    initPixels(myImage, GL_TEXTURE_2D);
    setWrap(GL_REPEAT);
    setFiltering(GL_LINEAR);
    if(makeMipMaps)
        createMipMaps();
}

Texture::Texture(Image* myCubemapImages[6], bool makeMipMaps) :
    m_texSlot(0),
    m_target(GL_TEXTURE_CUBE_MAP),
    m_width(myCubemapImages[0]->width),
    m_height(myCubemapImages[0]->height),
    m_dataType(GL_UNSIGNED_BYTE),
    m_hasMipMaps(makeMipMaps)
{
    glGenTextures(1, &m_texId);
    glBindTexture(m_target, m_texId);
    for(int i=0; i<6; ++i)
        initPixels(myCubemapImages[i], GL_TEXTURE_CUBE_MAP_POSITIVE_X + i);
    setWrap(GL_CLAMP_TO_EDGE);
    setFiltering(GL_LINEAR);
    if(makeMipMaps)
        createMipMaps();
}

Texture::Texture(Texture* tex, bool halfDim) :
    m_texSlot(0),
    m_target(tex->m_target),
    m_format(tex->m_format),
    m_internal_format(tex->m_internal_format),
    m_width(tex->m_width),
    m_height(tex->m_height),
    m_dataType(tex->m_dataType),
    m_hasMipMaps(false)
{
    glGenTextures(1, &m_texId);
    glBindTexture(m_target, m_texId);
    if(halfDim)
    {
        m_width /= 2;
        m_height /= 2;
    }
    glTexImage2D(m_target, 0, m_internal_format, m_width, m_height, 0, m_format, m_dataType, NULL);
}

Texture::~Texture()
{
    glDeleteTextures(1, &m_texId);
}

void Texture::initPixels(Image* myImage, GLenum target)
{
    switch(myImage->depth)
    {
    case 32:
        m_format = GL_RGBA;
        m_internal_format = GL_RGBA;
        break;
    case 24:
        m_format = GL_RGB;
        m_internal_format = GL_RGB;
        break;
    case 8:
        m_format = GL_RED;
        m_internal_format = GL_R8;
        break;
    }
    glTexImage2D(target, 0, m_internal_format, m_width, m_height, 0, m_format, m_dataType, myImage->pixels);
}

void Texture::setWrap(GLint wrap)
{
    setParameter(GL_TEXTURE_WRAP_S, wrap);
    setParameter(GL_TEXTURE_WRAP_T, wrap);
    setParameter(GL_TEXTURE_WRAP_R, wrap);
}

void Texture::setFiltering(GLint filter)
{
    setParameter(GL_TEXTURE_MIN_FILTER, filter);
    setParameter(GL_TEXTURE_MAG_FILTER, filter);
}

void Texture::setParameter(GLenum parameter, GLenum value)
{
    glTextureParameteri(m_texId, parameter, value);
}

void Texture::createMipMaps()
{
    m_hasMipMaps = true;
    glGenerateMipmap(m_target);
    setParameter(GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
}

void Texture::bind(int slot)
{
    if(slot >= 0)
        m_texSlot = slot;
    glActiveTexture(GL_TEXTURE0+m_texSlot);
    glBindTexture(m_target, m_texId);
}

void Texture::unbind()
{
    glActiveTexture(GL_TEXTURE0+m_texSlot);
    glBindTexture(m_target, 0);
}
